// Range v3 library
//
//  Copyright Eric Niebler 2014-present
//
//  Use, modification and distribution is subject to the
//  Boost Software License, Version 1.0. (See accompanying
//  file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt)
//
// Project home: https://github.com/ericniebler/range-v3

#include <range/v3/all.hpp>
#include "./simple_test.hpp"
#include "./test_utils.hpp"
#include "./test_iterators.hpp"

int main()
{
    return ::test_result();
}
